/* Window variables */
WINDOW *on_vol;


/* Common variables */
extern struct direntry      *directory;		 /* points to the directory */

extern int select_files(char *cur_dir);
extern int bmessage_box(char *s, int type);
extern void select_entry_engine(struct file_info *fi, struct file_entry *ce, 
			 char *fn, int printbox, int add, int sel_method);
